var DEFAULT_CALLBACKS, KEY_CODE;

KEY_CODE = {
  ESC: 27,
  TAB: 9,
  ENTER: 13,
  CTRL: 17,
  A: 65,
  P: 80,
  N: 78,
  LEFT: 37,
  UP: 38,
  RIGHT: 39,
  DOWN: 40,
  BACKSPACE: 8,
  SPACE: 32
};

// Functions set for handling and rendering the data.
// Others developers can override these methods to tweak At.js such as matcher.
// We can override them in `callbacks` settings.

// @mixin

// The context of these functions is `$.atwho.Controller` object and they are called in this sequences:

// [beforeSave, matcher, filter, remoteFilter, sorter, tplEvl, highlighter, beforeInsert, afterMatchFailed]

DEFAULT_CALLBACKS = {
  // It would be called to restructure the data before At.js invokes `Model#save` to save data
  // By default, At.js will convert it to a Hash Array.

  // @param data [Array] data to refacotor.
  // @return [Array] Data after refactor.
  beforeSave: function(data) {
    return Controller.arrayToDefaultHash(data);
  },
  // It would be called to match the `flag`.
  // It will match at start of line or after whitespace

  // @param flag [String] current `flag` ("@", etc)
  // @param subtext [String] Text from start to current caret position.
  // @param should_startWithSpace [boolean] accept white space as beginning of match.
  // @param acceptSpaceBar [boolean] accept a space bar in the center of match,
  //                                 so you can match a first and last name, for ex.

  // @return [String | null] Matched result.
  matcher: function(flag, subtext, should_startWithSpace, acceptSpaceBar) {
    var _a, _y, match, regexp, space;
    // escape RegExp
    flag = flag.replace(/[\-\[\]\/\{\}\(\)\*\+\?\.\\\^\$\|]/g, "\\$&");
    if (should_startWithSpace) {
      flag = '(?:^|\\s)' + flag;
    }
    // À
    _a = decodeURI("%C3%80");
    // ÿ
    _y = decodeURI("%C3%BF");
    space = acceptSpaceBar ? "\ " : "";
    regexp = new RegExp(`${flag}([A-Za-z${_a}-${_y}0-9_${space}\'\.\+\-]*)$|${flag}([^\\x00-\\xff]*)$`, 'gi');
    match = regexp.exec(subtext);
    if (match) {
      return match[2] || match[1];
    } else {
      return null;
    }
  },
  // ---------------------

  // Filter data by matched string.

  // @param query [String] Matched string.
  // @param data [Array] data list
  // @param searchKey [String] at char for searching.

  // @return [Array] result data.
  filter: function(query, data, searchKey) {
    var _results, i, item, len;
    // !!null #=> false; !!undefined #=> false; !!'' #=> false;
    _results = [];
    for (i = 0, len = data.length; i < len; i++) {
      item = data[i];
      if (~new String(item[searchKey]).toLowerCase().indexOf(query.toLowerCase())) {
        _results.push(item);
      }
    }
    return _results;
  },
  // If a function is given, At.js will invoke it if local filter can not find any data

  // @param params [String] matched query
  // @param callback [Function] callback to render page.
  remoteFilter: null,
  // remoteFilter: (query, callback) ->
  //   $.ajax url,
  //     data: params
  //     success: (data) ->
  //       callback(data)

  // Sorter data of course.

  // @param query [String] matched string
  // @param items [Array] data that was refactored
  // @param searchKey [String] at char to search

  // @return [Array] sorted data
  sorter: function(query, items, searchKey) {
    var _results, i, item, len;
    if (!query) {
      return items;
    }
    _results = [];
    for (i = 0, len = items.length; i < len; i++) {
      item = items[i];
      item.atwho_order = new String(item[searchKey]).toLowerCase().indexOf(query.toLowerCase());
      if (item.atwho_order > -1) {
        _results.push(item);
      }
    }
    return _results.sort(function(a, b) {
      return a.atwho_order - b.atwho_order;
    });
  },
  // Evaluate the template either as a string or as a function
  // this allows someone to pass in a set of data that needs a
  // different template for different data results

  // @param tpl [function] the template function or string
  // @param map [Hash] Data map to eval.
  tplEval: function(tpl, map) {
    var error, template;
    template = tpl;
    try {
      if (typeof tpl !== 'string') {
        template = tpl(map);
      }
      return template.replace(/\$\{([^\}]*)\}/g, function(tag, key, pos) {
        return map[key];
      });
    } catch (error1) {
      error = error1;
      return "";
    }
  },
  // Highlight the `matched query` string.

  // @param li [String] HTML String after eval.
  // @param query [String] matched query.

  // @return [String] highlighted string.
  highlighter: function(li, query) {
    var regexp;
    if (!query) {
      return li;
    }
    regexp = new RegExp(">\\s*([^\<]*?)(" + query.replace("+", "\\+") + ")([^\<]*)\\s*<", 'ig');
    return li.replace(regexp, function(str, $1, $2, $3) {
      return '> ' + $1 + '<strong>' + $2 + '</strong>' + $3 + ' <';
    });
  },
  // What to do before inserting item's value into inputor.

  // @param value [String] content to insert
  // @param $li [jQuery Object] the chosen item
  // @param e [event Object] from the user selection (keyDown or click)
  beforeInsert: function(value, $li, e) {
    return value;
  },
  // You can adjust the menu's offset here.

  // @param offset [Hash] offset will be applied to menu
  // beforeReposition: (offset) ->
  //   offset.left += 10
  //   offset.top += 10
  //   offset
  beforeReposition: function(offset) {
    return offset;
  },
  afterMatchFailed: function(at, el) {}
};

// At.js central contoller(searching, matching, evaluating and rendering.)
var App;

App = class App {
  // @param inputor [HTML DOM Object] `input` or `textarea`
  constructor(inputor) {
    this.currentFlag = null;
    this.controllers = {};
    this.aliasMaps = {};
    this.$inputor = $(inputor);
    this.setupRootElement();
    this.listen();
  }

  createContainer(doc) {
    var ref;
    if ((ref = this.$el) != null) {
      ref.remove();
    }
    return $(doc.body).append(this.$el = $("<div class='atwho-container'></div>"));
  }

  setupRootElement(iframe, asRoot = false) {
    var error;
    if (iframe) {
      this.window = iframe.contentWindow;
      this.document = iframe.contentDocument || this.window.document;
      this.iframe = iframe;
    } else {
      this.document = this.$inputor[0].ownerDocument;
      this.window = this.document.defaultView || this.document.parentWindow;
      try {
        this.iframe = this.window.frameElement;
      } catch (error1) {
        error = error1;
        this.iframe = null;
        if ($.fn.atwho.debug) {
          throw new Error(`iframe auto-discovery is failed.
Please use \`setIframe\` to set the target iframe manually.
${error}`);
        }
      }
    }
    return this.createContainer((this.iframeAsRoot = asRoot) ? this.document : document);
  }

  controller(at) {
    var c, current, currentFlag, ref;
    if (this.aliasMaps[at]) {
      current = this.controllers[this.aliasMaps[at]];
    } else {
      ref = this.controllers;
      for (currentFlag in ref) {
        c = ref[currentFlag];
        if (currentFlag === at) {
          current = c;
          break;
        }
      }
    }
    if (current) {
      return current;
    } else {
      return this.controllers[this.currentFlag];
    }
  }

  setContextFor(at) {
    this.currentFlag = at;
    return this;
  }

  // At.js can register multiple at char (flag) to every inputor such as "@" and ":"
  // Along with their own `settings` so that it works differently.
  // After register, we still can update their `settings` such as updating `data`

  // @param flag [String] at char (flag)
  // @param settings [Hash] the settings
  reg(flag, setting) {
    var base, controller;
    controller = (base = this.controllers)[flag] || (base[flag] = this.$inputor.is('[contentEditable]') ? new EditableController(this, flag) : new TextareaController(this, flag));
    if (setting.alias) {
      // TODO: it will produce rubbish alias map, reduse this.
      this.aliasMaps[setting.alias] = flag;
    }
    controller.init(setting);
    return this;
  }

  // binding jQuery events of `inputor`'s
  listen() {
    return this.$inputor.on('compositionstart', (e) => {
      var ref;
      if ((ref = this.controller()) != null) {
        ref.view.hide();
      }
      this.isComposing = true;
      return null;
    }).on('compositionend', (e) => {
      this.isComposing = false;
      setTimeout((e) => {
        return this.dispatch(e);
      });
      return null;
    }).on('keyup.atwhoInner', (e) => {
      return this.onKeyup(e);
    }).on('keydown.atwhoInner', (e) => {
      return this.onKeydown(e);
    }).on('blur.atwhoInner', (e) => {
      var c;
      if (c = this.controller()) {
        c.expectedQueryCBId = null;
        return c.view.hide(e, c.getOpt("displayTimeout"));
      }
    }).on('click.atwhoInner', (e) => {
      return this.dispatch(e);
    }).on('scroll.atwhoInner', (() => {
      var lastScrollTop;
      // make returned handler handle the very first call properly
      lastScrollTop = this.$inputor.scrollTop();
      return (e) => {
        var currentScrollTop, ref;
        currentScrollTop = e.target.scrollTop;
        if (lastScrollTop !== currentScrollTop) {
          if ((ref = this.controller()) != null) {
            ref.view.hide(e);
          }
        }
        lastScrollTop = currentScrollTop;
        return true; // ensure we don't stop bubbling
      };
    })());
  }

  shutdown() {
    var _, c, ref;
    ref = this.controllers;
    for (_ in ref) {
      c = ref[_];
      c.destroy();
      delete this.controllers[_];
    }
    this.$inputor.off('.atwhoInner');
    return this.$el.remove();
  }

  dispatch(e) {
    var _, c, ref, results;
    ref = this.controllers;
    results = [];
    for (_ in ref) {
      c = ref[_];
      results.push(c.lookUp(e));
    }
    return results;
  }

  onKeyup(e) {
    var ref;
    switch (e.keyCode) {
      case KEY_CODE.ESC:
        e.preventDefault();
        if ((ref = this.controller()) != null) {
          ref.view.hide();
        }
        break;
      case KEY_CODE.DOWN:
      case KEY_CODE.UP:
      case KEY_CODE.CTRL:
      case KEY_CODE.ENTER:
        $.noop();
        break;
      case KEY_CODE.P:
      case KEY_CODE.N:
        if (!e.ctrlKey) {
          this.dispatch(e);
        }
        break;
      default:
        this.dispatch(e);
    }
  }

  // coffeescript will return everywhere!!
  onKeydown(e) {
    var ref, view;
    // return if not (view = this.controller().view).visible()
    view = (ref = this.controller()) != null ? ref.view : void 0;
    if (!(view && view.visible())) {
      return;
    }
    switch (e.keyCode) {
      case KEY_CODE.ESC:
        e.preventDefault();
        view.hide(e);
        break;
      case KEY_CODE.UP:
        e.preventDefault();
        view.prev();
        break;
      case KEY_CODE.DOWN:
        e.preventDefault();
        view.next();
        break;
      case KEY_CODE.P:
        if (!e.ctrlKey) {
          return;
        }
        e.preventDefault();
        view.prev();
        break;
      case KEY_CODE.N:
        if (!e.ctrlKey) {
          return;
        }
        e.preventDefault();
        view.next();
        break;
      case KEY_CODE.TAB:
      case KEY_CODE.ENTER:
      case KEY_CODE.SPACE:
        if (!view.visible()) {
          return;
        }
        if (!this.controller().getOpt('spaceSelectsMatch') && e.keyCode === KEY_CODE.SPACE) {
          return;
        }
        if (!this.controller().getOpt('tabSelectsMatch') && e.keyCode === KEY_CODE.TAB) {
          return;
        }
        if (view.highlighted()) {
          e.preventDefault();
          view.choose(e);
        } else {
          view.hide(e);
        }
        break;
      default:
        $.noop();
    }
  }

};

var Controller;

Controller = class Controller {
  uid() {
    return (Math.random().toString(16) + "000000000").substr(2, 8) + (new Date().getTime());
  }

  constructor(app, at1) {
    this.app = app;
    this.at = at1;
    this.$inputor = this.app.$inputor;
    this.id = this.$inputor[0].id || this.uid();
    this.expectedQueryCBId = null;
    this.setting = null;
    this.query = null;
    this.pos = 0;
    this.range = null;
    if ((this.$el = $(`#atwho-ground-${this.id}`, this.app.$el)).length === 0) {
      this.app.$el.append(this.$el = $(`<div id='atwho-ground-${this.id}'></div>`));
    }
    this.model = new Model(this);
    this.view = new View(this);
  }

  init(setting) {
    this.setting = $.extend({}, this.setting || $.fn.atwho.default, setting);
    this.view.init();
    return this.model.reload(this.setting.data);
  }

  destroy() {
    this.trigger('beforeDestroy');
    this.model.destroy();
    this.view.destroy();
    return this.$el.remove();
  }

  callDefault(funcName, ...args) {
    var error;
    try {
      return DEFAULT_CALLBACKS[funcName].apply(this, args);
    } catch (error1) {
      error = error1;
      return $.error(`${error} Or maybe At.js doesn't have function ${funcName}`);
    }
  }

  // Delegate custom `jQueryEvent` to the inputor
  // This function will add `atwho` as namespace to every jQuery event
  // and pass current context as the last param to it.

  // @example
  //   this.trigger "roll_n_rock", [1,2,3,4]

  //   $inputor.on "rool_n_rock", (e, one, two, three, four) ->
  //     console.log one, two, three, four

  // @param name [String] Event name
  // @param data [Array] data to callback
  trigger(name, data = []) {
    var alias, eventName;
    data.push(this);
    alias = this.getOpt('alias');
    eventName = alias ? `${name}-${alias}.atwho` : `${name}.atwho`;
    return this.$inputor.trigger(eventName, data);
  }

  // Get callback either in settings which was set by plugin user or in default callbacks list.

  // @param funcName [String] callback's name
  // @return [Function] The callback.
  callbacks(funcName) {
    return this.getOpt("callbacks")[funcName] || DEFAULT_CALLBACKS[funcName];
  }

  // Because different registered at chars have different settings.
  // so we should give their own for them.

  // @param at [String] setting's at name
  // @param default_value [?] return this if nothing is returned from current settings.
  // @return [?] setting's value
  getOpt(at, default_value) {
    var e;
    try {
      return this.setting[at];
    } catch (error1) {
      e = error1;
      return null;
    }
  }

  insertContentFor($li) {
    var data, tpl;
    tpl = this.getOpt('insertTpl');
    data = $.extend({}, $li.data('item-data'), {
      'atwho-at': this.at
    });
    return this.callbacks("tplEval").call(this, tpl, data, "onInsert");
  }

  // Render list view

  // @param data [Array] The data
  renderView(data) {
    var searchKey;
    searchKey = this.getOpt("searchKey");
    data = this.callbacks("sorter").call(this, this.query.text, data.slice(0, 1001), searchKey);
    return this.view.render(data.slice(0, this.getOpt('limit')));
  }

  static arrayToDefaultHash(data) {
    var i, item, len, results;
    if (!$.isArray(data)) {
      return data;
    }
    results = [];
    for (i = 0, len = data.length; i < len; i++) {
      item = data[i];
      if ($.isPlainObject(item)) {
        results.push(item);
      } else {
        results.push({
          name: item
        });
      }
    }
    return results;
  }

  // Searching!
  lookUp(e) {
    var query, wait;
    if (e && e.type === 'click' && !this.getOpt('lookUpOnClick')) {
      return;
    }
    if (this.getOpt('suspendOnComposing') && this.app.isComposing) {
      return;
    }
    query = this.catchQuery(e);
    if (!query) {
      this.expectedQueryCBId = null;
      return query;
    }
    this.app.setContextFor(this.at);
    if (wait = this.getOpt('delay')) {
      this._delayLookUp(query, wait);
    } else {
      this._lookUp(query);
    }
    return query;
  }

  _delayLookUp(query, wait) {
    var now, remaining;
    now = Date.now ? Date.now() : new Date().getTime();
    this.previousCallTime || (this.previousCallTime = now);
    remaining = wait - (now - this.previousCallTime);
    if ((0 < remaining && remaining < wait)) {
      this.previousCallTime = now;
      this._stopDelayedCall();
      return this.delayedCallTimeout = setTimeout(() => {
        this.previousCallTime = 0;
        this.delayedCallTimeout = null;
        return this._lookUp(query);
      }, wait);
    } else {
      this._stopDelayedCall();
      if (this.previousCallTime !== now) {
        this.previousCallTime = 0;
      }
      return this._lookUp(query);
    }
  }

  _stopDelayedCall() {
    if (this.delayedCallTimeout) {
      clearTimeout(this.delayedCallTimeout);
      return this.delayedCallTimeout = null;
    }
  }

  _generateQueryCBId() {
    return {};
  }

  _lookUp(query) {
    var _callback;
    _callback = function(queryCBId, data) {
      // ensure only the latest instance of this function perform actions
      if (queryCBId !== this.expectedQueryCBId) {
        return;
      }
      if (data && data.length > 0) {
        return this.renderView(this.constructor.arrayToDefaultHash(data));
      } else {
        return this.view.hide();
      }
    };
    this.expectedQueryCBId = this._generateQueryCBId();
    return this.model.query(query.text, $.proxy(_callback, this, this.expectedQueryCBId));
  }

};

var TextareaController;

TextareaController = class TextareaController extends Controller {
  // Catch query string behind the at char

  // @return [Hash] Info of the query. Look likes this: {'text': "hello", 'headPos': 0, 'endPos': 0}
  catchQuery() {
    var caretPos, content, end, isString, query, start, subtext;
    content = this.$inputor.val();
    caretPos = this.$inputor.caret('pos', {
      iframe: this.app.iframe
    });
    subtext = content.slice(0, caretPos);
    query = this.callbacks("matcher").call(this, this.at, subtext, this.getOpt('startWithSpace'), this.getOpt("acceptSpaceBar"));
    isString = typeof query === 'string';
    if (isString && query.length < this.getOpt('minLen', 0)) {
      return;
    }
    if (isString && query.length <= this.getOpt('maxLen', 20)) {
      start = caretPos - query.length;
      end = start + query.length;
      this.pos = start;
      query = {
        'text': query,
        'headPos': start,
        'endPos': end
      };
      this.trigger("matched", [this.at, query.text]);
    } else {
      query = null;
      this.view.hide();
    }
    return this.query = query;
  }

  // Get offset of current at char(`flag`)

  // @return [Hash] the offset which look likes this: {top: y, left: x, bottom: bottom}
  rect() {
    var c, iframeOffset, scaleBottom;
    if (!(c = this.$inputor.caret('offset', this.pos - 1, {
      iframe: this.app.iframe
    }))) {
      return;
    }
    if (this.app.iframe && !this.app.iframeAsRoot) {
      iframeOffset = $(this.app.iframe).offset();
      c.left += iframeOffset.left;
      c.top += iframeOffset.top;
    }
    scaleBottom = this.app.document.selection ? 0 : 2;
    return {
      left: c.left,
      top: c.top,
      bottom: c.top + c.height + scaleBottom
    };
  }

  // Insert value of `data-value` attribute of chosen item into inputor

  // @param content [String] string to insert
  insert(content, $li) {
    var $inputor, source, startStr, suffix, text;
    $inputor = this.$inputor;
    source = $inputor.val();
    startStr = source.slice(0, Math.max(this.query.headPos - this.at.length, 0));
    suffix = (suffix = this.getOpt('suffix')) === "" ? suffix : suffix || " ";
    content += suffix;
    text = `${startStr}${content}${source.slice(this.query['endPos'] || 0)}`;
    $inputor.val(text);
    $inputor.caret('pos', startStr.length + content.length, {
      iframe: this.app.iframe
    });
    if (!$inputor.is(':focus')) {
      $inputor.focus();
    }
    return $inputor.change();
  }

};

var EditableController;

EditableController = class EditableController extends Controller {
  _getRange() {
    var sel;
    sel = this.app.window.getSelection();
    if (sel.rangeCount > 0) {
      return sel.getRangeAt(0);
    }
  }

  _setRange(position, node, range = this._getRange()) {
    if (!(range && node)) {
      return;
    }
    node = $(node)[0];
    if (position === 'after') {
      range.setEndAfter(node);
      range.setStartAfter(node);
    } else {
      range.setEndBefore(node);
      range.setStartBefore(node);
    }
    range.collapse(false);
    return this._clearRange(range);
  }

  _clearRange(range = this._getRange()) {
    var sel;
    sel = this.app.window.getSelection();
    if (this.ctrl_a_pressed == null) {
      sel.removeAllRanges();
      return sel.addRange(range);
    }
  }

  _movingEvent(e) {
    var ref;
    return e.type === 'click' || ((ref = e.which) === KEY_CODE.RIGHT || ref === KEY_CODE.LEFT || ref === KEY_CODE.UP || ref === KEY_CODE.DOWN);
  }

  _unwrap(node) {
    var next;
    node = $(node).unwrap().get(0);
    if ((next = node.nextSibling) && next.nodeValue) {
      node.nodeValue += next.nodeValue;
      $(next).remove();
    }
    return node;
  }

  catchQuery(e) {
    var $inserted, $query, _range, index, inserted, isString, lastNode, matched, offset, query, query_content, range;
    if (!(range = this._getRange())) {
      return;
    }
    if (!range.collapsed) {
      return;
    }
    if (e.which === KEY_CODE.ENTER) {
      ($query = $(range.startContainer).closest('.atwho-query')).contents().unwrap();
      if ($query.is(':empty')) {
        $query.remove();
      }
      ($query = $(".atwho-query", this.app.document)).text($query.text()).contents().last().unwrap();
      this._clearRange();
      return;
    }
    // absorb range
    // The range at the end of an element is not inside in firefox but not others browsers including IE.
    // To normolize them, we have to move the range inside the element while deleting content or moving caret right after .atwho-inserted
    if (/firefox/i.test(navigator.userAgent)) {
      if ($(range.startContainer).is(this.$inputor)) {
        this._clearRange();
        return;
      }
      if (e.which === KEY_CODE.BACKSPACE && range.startContainer.nodeType === document.ELEMENT_NODE && (offset = range.startOffset - 1) >= 0) {
        _range = range.cloneRange();
        _range.setStart(range.startContainer, offset);
        if ($(_range.cloneContents()).contents().last().is('.atwho-inserted')) {
          inserted = $(range.startContainer).contents().get(offset);
          this._setRange('after', $(inserted).contents().last());
        }
      } else if (e.which === KEY_CODE.LEFT && range.startContainer.nodeType === document.TEXT_NODE) {
        $inserted = $(range.startContainer.previousSibling);
        if ($inserted.is('.atwho-inserted') && range.startOffset === 0) {
          this._setRange('after', $inserted.contents().last());
        }
      }
    }
    // modifying inserted element
    $(range.startContainer).closest('.atwho-inserted').addClass('atwho-query').siblings().removeClass('atwho-query');
    if (($query = $(".atwho-query", this.app.document)).length > 0 && $query.is(':empty') && $query.text().length === 0) {
      $query.remove();
    }
    if (!this._movingEvent(e)) {
      $query.removeClass('atwho-inserted');
    }
    if ($query.length > 0) {
      switch (e.which) {
        case KEY_CODE.LEFT:
          this._setRange('before', $query.get(0), range);
          $query.removeClass('atwho-query');
          return;
        case KEY_CODE.RIGHT:
          this._setRange('after', $query.get(0).nextSibling, range);
          $query.removeClass('atwho-query');
          return;
      }
    }
    // matching
    if ($query.length > 0 && (query_content = $query.attr('data-atwho-at-query'))) {
      $query.empty().html(query_content).attr('data-atwho-at-query', null);
      this._setRange('after', $query.get(0), range);
    }
    _range = range.cloneRange();
    _range.setStart(range.startContainer, 0);
    matched = this.callbacks("matcher").call(this, this.at, _range.toString(), this.getOpt('startWithSpace'), this.getOpt("acceptSpaceBar"));
    isString = typeof matched === 'string';
    // wrapping query with .atwho-query
    if ($query.length === 0 && isString && (index = range.startOffset - this.at.length - matched.length) >= 0) {
      range.setStart(range.startContainer, index);
      $query = $('<span/>', this.app.document).attr(this.getOpt("editableAtwhoQueryAttrs")).addClass('atwho-query');
      range.surroundContents($query.get(0));
      lastNode = $query.contents().last().get(0);
      if (lastNode) {
        if (/firefox/i.test(navigator.userAgent)) {
          range.setStart(lastNode, lastNode.length);
          range.setEnd(lastNode, lastNode.length);
          this._clearRange(range);
        } else {
          this._setRange('after', lastNode, range);
        }
      }
    }
    if (isString && matched.length < this.getOpt('minLen', 0)) {
      return;
    }
    // handle the matched result
    if (isString && matched.length <= this.getOpt('maxLen', 20)) {
      query = {
        text: matched,
        el: $query
      };
      this.trigger("matched", [this.at, query.text]);
      return this.query = query;
    } else {
      this.view.hide();
      this.query = {
        el: $query
      };
      if ($query.text().indexOf(this.at) >= 0) {
        if (this._movingEvent(e) && $query.hasClass('atwho-inserted')) {
          $query.removeClass('atwho-query');
        } else if (false !== this.callbacks('afterMatchFailed').call(this, this.at, $query)) {
          this._setRange("after", this._unwrap($query.text($query.text()).contents().first()));
        }
      }
      return null;
    }
  }

  // Get offset of current at char(`flag`)

  // @return [Hash] the offset which look likes this: {top: y, left: x, bottom: bottom}
  rect() {
    var $iframe, iframeOffset, rect;
    rect = this.query.el.offset();
    // do not use {top: 0, left: 0} from jQuery when element is hidden
    // happens every other time the menu is displayed on click in contenteditable
    if (!(rect && this.query.el[0].getClientRects().length)) {
      return;
    }
    if (this.app.iframe && !this.app.iframeAsRoot) {
      iframeOffset = ($iframe = $(this.app.iframe)).offset();
      rect.left += iframeOffset.left - this.$inputor.scrollLeft();
      rect.top += iframeOffset.top - this.$inputor.scrollTop();
    }
    rect.bottom = rect.top + this.query.el.height();
    return rect;
  }

  // Insert value of `data-value` attribute of chosen item into inputor

  // @param content [String] string to insert
  insert(content, $li) {
    var data, overrides, range, suffix, suffixNode;
    if (!this.$inputor.is(':focus')) {
      this.$inputor.focus();
    }
    overrides = this.getOpt('functionOverrides');
    if (overrides.insert) {
      return overrides.insert.call(this, content, $li);
    }
    suffix = (suffix = this.getOpt('suffix')) === "" ? suffix : suffix || "\u00A0";
    data = $li.data('item-data');
    this.query.el.removeClass('atwho-query').addClass('atwho-inserted').html(content).attr('data-atwho-at-query', "" + data['atwho-at'] + this.query.text).attr('contenteditable', "false");
    if (range = this._getRange()) {
      if (this.query.el.length) {
        range.setEndAfter(this.query.el[0]);
      }
      range.collapse(false);
      range.insertNode(suffixNode = this.app.document.createTextNode("" + suffix));
      this._setRange('after', suffixNode, range);
    }
    if (!this.$inputor.is(':focus')) {
      this.$inputor.focus();
    }
    return this.$inputor.change();
  }

};

// Class to process data
var Model;

Model = class Model {
  constructor(context) {
    this.context = context;
    this.at = this.context.at;
    // NOTE: bind data storage to inputor maybe App class can handle it.
    this.storage = this.context.$inputor;
  }

  destroy() {
    return this.storage.data(this.at, null);
  }

  saved() {
    return this.fetch() > 0;
  }

  // fetch data from storage by query.
  // will invoke `callback` to return data

  // @param query [String] catched string for searching
  // @param callback [Function] for receiving data
  query(query, callback) {
    var _remoteFilter, data, searchKey;
    data = this.fetch();
    searchKey = this.context.getOpt("searchKey");
    data = this.context.callbacks('filter').call(this.context, query, data, searchKey) || [];
    _remoteFilter = this.context.callbacks('remoteFilter');
    if (data.length > 0 || (!_remoteFilter && data.length === 0)) {
      return callback(data);
    } else {
      return _remoteFilter.call(this.context, query, callback);
    }
  }

  // get or set current data which would be shown on the list view.

  // @param data [Array] set data
  // @return [Array|undefined] current data that are showing on the list view.
  fetch() {
    return this.storage.data(this.at) || [];
  }

  // save special flag's data to storage

  // @param data [Array] data to save
  save(data) {
    return this.storage.data(this.at, this.context.callbacks("beforeSave").call(this.context, data || []));
  }

  // load data. It wouldn't load for a second time if it has been loaded.

  // @param data [Array] data to load
  load(data) {
    if (!(this.saved() || !data)) {
      return this._load(data);
    }
  }

  reload(data) {
    return this._load(data);
  }

  // load data from local or remote with callback

  // @param data [Array|String] data to load.
  _load(data) {
    if (typeof data === "string") {
      return $.ajax(data, {
        dataType: "json"
      }).done((data) => {
        return this.save(data);
      });
    } else {
      return this.save(data);
    }
  }

};

// View class to control how At.js's view showing.
// All classes share the same DOM view.
var View;

View = class View {
  // @param controller [Object] The Controller.
  constructor(context) {
    this.context = context;
    this.$el = $("<div class='atwho-view'><ul class='atwho-view-ul'></ul></div>");
    this.$elUl = this.$el.children();
    this.timeoutID = null;
    // create HTML DOM of list view if it does not exist
    this.context.$el.append(this.$el);
    this.bindEvent();
  }

  init() {
    var header_tpl, id;
    id = this.context.getOpt("alias") || this.context.at.charCodeAt(0);
    header_tpl = this.context.getOpt("headerTpl");
    if (header_tpl && this.$el.children().length === 1) {
      this.$el.prepend(header_tpl);
    }
    return this.$el.attr({
      'id': `at-view-${id}`
    });
  }

  destroy() {
    return this.$el.remove();
  }

  bindEvent() {
    var $menu, lastCoordX, lastCoordY;
    $menu = this.$el.find('ul');
    lastCoordX = 0;
    lastCoordY = 0;
    return $menu.on('mousemove.atwho-view', 'li', (e) => {
      var $cur;
      // If the mouse hasn't actually moved then exit.
      if (lastCoordX === e.clientX && lastCoordY === e.clientY) {
        return;
      }
      lastCoordX = e.clientX;
      lastCoordY = e.clientY;
      $cur = $(e.currentTarget);
      if ($cur.hasClass('cur')) {
        return;
      }
      $menu.find('.cur').removeClass('cur');
      return $cur.addClass('cur');
    }).on('click.atwho-view', 'li', (e) => {
      $menu.find('.cur').removeClass('cur');
      $(e.currentTarget).addClass('cur');
      this.choose(e);
      return e.preventDefault();
    });
  }

  // Check if view is visible

  // @return [Boolean]
  visible() {
    return $.expr.filters.visible(this.$el[0]);
  }

  highlighted() {
    return this.$el.find(".cur").length > 0;
  }

  choose(e) {
    var $li, content;
    if (($li = this.$el.find(".cur")).length) {
      content = this.context.insertContentFor($li);
      this.context._stopDelayedCall();
      this.context.insert(this.context.callbacks("beforeInsert").call(this.context, content, $li, e), $li);
      this.context.trigger("inserted", [$li, e]);
      this.hide(e);
    }
    if (this.context.getOpt("hideWithoutSuffix")) {
      return this.stopShowing = true;
    }
  }

  reposition(rect) {
    var _window, offset, overflowOffset, ref;
    _window = this.context.app.iframeAsRoot ? this.context.app.window : window;
    if (rect.bottom + this.$el.height() - $(_window).scrollTop() > $(_window).height()) {
      rect.bottom = rect.top - this.$el.height();
    }
    if (rect.left > (overflowOffset = $(_window).width() - this.$el.width() - 5)) {
      rect.left = overflowOffset;
    }
    offset = {
      left: rect.left,
      top: rect.bottom
    };
    if ((ref = this.context.callbacks("beforeReposition")) != null) {
      ref.call(this.context, offset);
    }
    this.$el.offset(offset);
    return this.context.trigger("reposition", [offset]);
  }

  next() {
    var cur, next, nextEl, offset;
    cur = this.$el.find('.cur').removeClass('cur');
    next = cur.next();
    if (!next.length) {
      next = this.$el.find('li:first');
    }
    next.addClass('cur');
    nextEl = next[0];
    offset = nextEl.offsetTop + nextEl.offsetHeight + (nextEl.nextSibling ? nextEl.nextSibling.offsetHeight : 0);
    return this.scrollTop(Math.max(0, offset - this.$el.height()));
  }

  prev() {
    var cur, offset, prev, prevEl;
    cur = this.$el.find('.cur').removeClass('cur');
    prev = cur.prev();
    if (!prev.length) {
      prev = this.$el.find('li:last');
    }
    prev.addClass('cur');
    prevEl = prev[0];
    offset = prevEl.offsetTop + prevEl.offsetHeight + (prevEl.nextSibling ? prevEl.nextSibling.offsetHeight : 0);
    return this.scrollTop(Math.max(0, offset - this.$el.height()));
  }

  scrollTop(scrollTop) {
    var scrollDuration;
    scrollDuration = this.context.getOpt('scrollDuration');
    if (scrollDuration) {
      return this.$elUl.animate({
        scrollTop: scrollTop
      }, scrollDuration);
    } else {
      return this.$elUl.scrollTop(scrollTop);
    }
  }

  show() {
    var rect;
    if (this.stopShowing) {
      this.stopShowing = false;
      return;
    }
    if (!this.visible()) {
      this.$el.show();
      this.$el.scrollTop(0);
      this.context.trigger('shown');
    }
    if (rect = this.context.rect()) {
      return this.reposition(rect);
    }
  }

  hide(e, time) {
    var callback;
    if (!this.visible()) {
      return;
    }
    if (isNaN(time)) {
      this.$el.hide();
      return this.context.trigger('hidden', [e]);
    } else {
      callback = () => {
        return this.hide();
      };
      clearTimeout(this.timeoutID);
      return this.timeoutID = setTimeout(callback, time);
    }
  }

  // render list view
  render(list) {
    var $li, $ul, i, item, len, li, tpl;
    if (!($.isArray(list) && list.length > 0)) {
      this.hide();
      return;
    }
    this.$el.find('ul').empty();
    $ul = this.$el.find('ul');
    tpl = this.context.getOpt('displayTpl');
    for (i = 0, len = list.length; i < len; i++) {
      item = list[i];
      item = $.extend({}, item, {
        'atwho-at': this.context.at
      });
      li = this.context.callbacks("tplEval").call(this.context, tpl, item, "onDisplay");
      $li = $(this.context.callbacks("highlighter").call(this.context, li, this.context.query.text));
      $li.data("item-data", item);
      $ul.append($li);
    }
    this.show();
    if (this.context.getOpt('highlightFirst')) {
      return $ul.find("li:first").addClass("cur");
    }
  }

};

var Api;

Api = {
  // load a flag's data

  // @params at[String] the flag
  // @params data [Array] data to storage.
  load: function(at, data) {
    var c;
    if (c = this.controller(at)) {
      return c.model.load(data);
    }
  },
  isSelecting: function() {
    var ref;
    return !!((ref = this.controller()) != null ? ref.view.visible() : void 0);
  },
  hide: function() {
    var ref;
    return (ref = this.controller()) != null ? ref.view.hide() : void 0;
  },
  reposition: function() {
    var c;
    if (c = this.controller()) {
      return c.view.reposition(c.rect());
    }
  },
  setIframe: function(iframe, asRoot) {
    this.setupRootElement(iframe, asRoot);
    return null;
  },
  run: function() {
    return this.dispatch();
  },
  destroy: function() {
    this.shutdown();
    return this.$inputor.data('atwho', null);
  }
};

$.fn.atwho = function(method) {
  var _args, result;
  _args = arguments;
  result = null;
  this.filter('textarea, input, [contenteditable=""], [contenteditable=true]').each(function() {
    var $this, app;
    if (!(app = ($this = $(this)).data("atwho"))) {
      $this.data('atwho', (app = new App(this)));
    }
    if (typeof method === 'object' || !method) {
      return app.reg(method.at, method);
    } else if (Api[method] && app) {
      return result = Api[method].apply(app, Array.prototype.slice.call(_args, 1));
    } else {
      return $.error(`Method ${method} does not exist on jQuery.atwho`);
    }
  });
  if (result != null) {
    return result;
  } else {
    return this;
  }
};

$.fn.atwho.default = {
  at: void 0,
  alias: void 0,
  data: null,
  displayTpl: "<li>${name}</li>",
  insertTpl: "${atwho-at}${name}",
  headerTpl: null,
  callbacks: DEFAULT_CALLBACKS,
  functionOverrides: {},
  searchKey: "name",
  suffix: void 0,
  hideWithoutSuffix: false,
  startWithSpace: true,
  acceptSpaceBar: false,
  highlightFirst: true,
  limit: 5,
  maxLen: 20,
  minLen: 0,
  displayTimeout: 300,
  delay: null,
  spaceSelectsMatch: false,
  tabSelectsMatch: true,
  editableAtwhoQueryAttrs: {},
  scrollDuration: 150,
  suspendOnComposing: true,
  lookUpOnClick: true
};

$.fn.atwho.debug = false;
